#ifndef __NC_APP__
#define __NC_APP__

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <allegro5/allegro.h>
#include <allegro5/allegro_font.h>
#include <allegro5/allegro_primitives.h>
#include <allegro5/allegro_native_dialog.h>
#include <allegro5/allegro_ttf.h>
#include <allegro5/allegro_image.h>
#include <allegro5/allegro_audio.h>
#include <allegro5/allegro_acodec.h>

#include "../render/background.h"
#include "../sprite/sprite.h"
#include "../point/point.h"
#include "../player/player.h"
#include "../enemy/leaf.h"
#include "../enemy/ball.h"
#include "../enemy/nbomb.h"
#include "../enemy/bigleaf.h"
#include "../enemy/orb.h"
#include "../enemy/flame.h"
#include "../enemy/bigorb.h"
#include "../enemy/slime.h"
#include "../enemy/flame2.h"
#include "../enemy/bulletb.h"
#include "../enemy/flyguy.h"

#define MACRO_POP_ERROR_MESSAGE(text,id)  al_show_native_message_box(display,"Error occurred!",id, text ,NULL,ALLEGRO_MESSAGEBOX_ERROR); return -1;
#define MACRO_LOAD_BITMAP(bmp,fname) bmp = al_load_bitmap(fname); if(!bmp) { al_destroy_display(display); al_show_native_message_box(NULL,"Error occurred! A Bitmap is missing!","Error E_BMP!", fname ,NULL,ALLEGRO_MESSAGEBOX_ERROR); return -1; }
#define MACRO_LOAD_SAMPLE(sample,fname) sample = al_load_sample(fname); if(!sample) { al_show_native_message_box(NULL,"Error E_AUDIO!","Error occurred! An audio file is missing!", fname ,NULL,ALLEGRO_MESSAGEBOX_ERROR); return -1; }

// Enum
typedef enum { SCENE_CREATED_BY = 0, SCENE_TITLE = 1, SCENE_STAGE = 2, SCENE_GAME = 3, SCENE_PAUSE = 4, SCENE_GAMEOVER = 5, SCENE_CLEAR = 6} sceneType;

// System variables
ALLEGRO_DISPLAY *display;
ALLEGRO_TIMER *fps_timer;
ALLEGRO_EVENT_QUEUE *main_events;
ALLEGRO_EVENT mainloop_event;
ALLEGRO_BITMAP * canvas, * canvas_snapshot;
ALLEGRO_FONT * font;

// Filter
ALLEGRO_BITMAP * scanlines;

//Bitmaps
ALLEGRO_BITMAP * bg1,
               * surface,
               * pl_bmp,
               * bullet_bmp,
               * bomb_bmp,
               * life,
               * leaf_bmp,
               * ball_bmp,
               * nbomb_bmp,
               * stage1_thumb,
               * title,
               * gameover_bmp,
               * bigleaf_bmp,
               * bg2,
               * surface2,
               * stage2_thumb,
               * nbomb2_bmp,
               * orb_bmp,
               * flame_bmp,
               * bigorb_bmp,
               * bg3,
               * surface3,
               * slime_bmp,
               * stage3_thumb,
               * flame2_bmp,
               * bulletb_bmp,
               * flyguy_bmp,
               * completed,
               * creator,
               * bg4,
               * surface4,
               * cloud_bmp,
               * stage4_thumb;

// Samples

ALLEGRO_SAMPLE * sChoose,
               * sSelect,
               * sThrow,
               * sDrop,
               * sHit,
               * sPropeller,
               * sHurt;

// Input
bool keymap[256];
bool joystick_enabled;

// Application functions
int app_init();
void app_destroy();
void app_run();
bool app_mainloop();
void app_paint(ALLEGRO_BITMAP*);
void app_events();
void app_recreate();

// Game objects
objPlayer player;
leaf leaves[10];
ball balls[10];
nbomb nbombs[10];
orb orbs[10];
flame flames[10];
slime slimes[10];
flame2 flame2s[10];
bulletb bulletbs[10];
bigleaf bigleaves[10];
bigorb bigorbs[10];
flyguy flyguys[10];
bigleaf boss_bl;
bigorb boss_bo;
flyguy boss_fg;
boostEffect boosts[10];
bullet bullets[50];
int current_boost,
    current_bullet,
    current_nbomb,
    current_orb,
    current_flame2;

// Game logic
sceneType scene;
unsigned int sceneTimer, secondaryTimer;
uint8_t sceneMode;
int8_t cursorPos;
long int score, last_score;
int lives, stage, multiplier, multiplier_timer, hp_up_meter;
uint8_t num_type1_enemies,
    num_type2_enemies,
    num_type3_enemies,
    num_type4_enemies,
    tries;
float clouds[3];
bool reborn;

#endif // __NC_APP__
