static unsigned char collision[(64 * 30) / 8];
#define MAX_POWERUPS_PER_ROOM 2
static unsigned char powerUpLocations[MAX_POWERUPS_PER_ROOM * 2];
static unsigned char levelIndex = 0xFF;
static unsigned char roomIndex;
static unsigned char keyLocationX;
static unsigned char keyLocationY;
static unsigned int pickedUpPowerups;

// note: each levels can have 8 rooms, have a byte to check which room is cleared of powerups, max 1 powerup per room

unsigned char isSolid(unsigned char _x, unsigned char _y)
{
	// each row has 4 bytes
	t0 = collision[(_y * 8) + _x / 8];
	
	return (t0 >> (_x % 8)) & 0x01;
}

void __fastcall__ switchTile(void)
{
	#define tile t0
	switch (levelIndex)
	{
		case 0:
			if ((tile >= 0x60 && tile <= 0x67) ||
				(tile >= 0x70 && tile <= 0x77) ||
				(tile >= 0x80 && tile <= 0x87))
			{
				tile += 8;
			}
			break;
		case 1:
			if ((tile >= 0x68 && tile <= 0x6F) ||
				(tile >= 0x78 && tile <= 0x7F) ||
				(tile >= 0x88 && tile <= 0x8F))
			{
				tile -= 8;
			}
			break;
	}
	vram_adr(ti1);
	vram_put(tile);
	#undef tile
}

#define powIndex t3
#define colIndex t2
#define colByte t1
#define tile t0
#define enemyIndex t4
extern unsigned char enemyType[];
extern unsigned char enemyAnim[];
extern unsigned char enemyVar[];
extern unsigned char enemyVar2[];
extern unsigned char enemyStun[];
extern int enemyX[];
extern int enemyY[];
extern unsigned char enemyHealth[];
void __fastcall__ processTile(void)
{
	if (tile == 0xE3) // the key
	{
		if (!hasKey)
		{
			keyLocationX = x;
			keyLocationY = y;
		}
		vram_adr(ti1);
		vram_put(0);
	}
	else if (tile >= 0xF0) // it's a power up
	{
		if (!(pickedUpPowerups & (0x0001 << ((roomIndex * 2) + powIndex))))
		{
			powerUpLocations[powIndex * 2] = tile == 0xF0 ? x : (x | 0x80);
			powerUpLocations[powIndex * 2 + 1] = y;
		}
		++powIndex;
		vram_adr(ti1);
		vram_put(0);
	}
	else if (tile >= 0x0D && tile <= 0x0F) // it's an enemy spawner
	{
		enemyType[enemyIndex] = (tile - 0x0D) + 1;
		enemyX[enemyIndex] = (x + 1) * 8 * SUB_PIXELS;
		enemyY[enemyIndex] = (y + 1) * 8 * SUB_PIXELS - 8;
		enemyVar[enemyIndex] = 0;
		switch (enemyType[enemyIndex])
		{
			case ROTNIP: enemyHealth[enemyIndex] = 3; break;
			case FLOPFLOP: enemyHealth[enemyIndex] = 1; break;
			case FISHBUN: enemyHealth[enemyIndex] = 2; break;
		}
		vram_adr(ti1);
		vram_put(0);
		++enemyIndex;
	}
	else if (tile >= 0x60) // if it's a solid
	{
		colByte |= 1 << i;
	}
}

void __fastcall__ processRoom(void)
{
	// clear power up locations
	oam_clear();
	for (i = 0; i < MAX_POWERUPS_PER_ROOM * 2; ++i)
		powerUpLocations[i] = 0;
	for (i = 0; i < MAX_ENEMIES; ++i)
	{
		enemyType[i] = 0;
		enemyVar[i] = 0;
		enemyVar2[i] = 0;
		enemyStun[i] = 0;
	}
	keyLocationX = 0;
	keyLocationY = 0;

	colIndex = 0;
	powIndex = 0;
	enemyIndex = 0;
	for(ti1 = 0x2000; ti1 < (0x2000 + (32 * 30));)
	{
		colByte = 0; // collision byte
		for (i = 0; i < 8; ++i) // create the byte
		{
			vram_read(&tile, ti1, 1);
			
			switchTile(); // switches tiles if they are in the wrong tileset
			
			// get the position of this tile
			x = (ti1 - 0x2000) % 32;
			y = (ti1 - 0x2000) / 32;
			
			processTile();
			
			++ti1;
		}
		collision[colIndex++] = colByte; // commit the byte
		if ((colIndex % 4) == 0)
			colIndex += 4;
	}
	
	colIndex = 4; // collision index
	for(ti1 = 0x2400; ti1 < (0x2400 + (32 * 30));)
	{
		colByte = 0; // collision byte
		for (i = 0; i < 8; ++i) // create the byte
		{
			vram_read(&tile, ti1, 1);
			
			switchTile(); // switches tiles if they are in the wrong tileset
			
			// get the position of this tile
			x = 32 + ((ti1 - 0x2400) % 32);
			y = (ti1 - 0x2400) / 32;
			
			processTile();
			
			++ti1;
		}
		collision[colIndex++] = colByte; // commit the byte
		if ((colIndex % 4) == 0)
			colIndex += 4;
	}
}
#undef tile
#undef colByte
#undef colIndex
#undef powIndex
#undef enemyIndex

extern unsigned char health;
void playerInit(unsigned char x, unsigned char y);
void loadLevel(unsigned char index)
{
	fadeOut();
	oam_clear();
	
	if (index != levelIndex)
	{
		// play music
		switch (index)
		{
			default:
			case 0:
				pal_bg(pal_greenforest);
				pal_spr(pal_sprites_greenforest);
				music_play(music_stage1); 
				break;
			case 1:
				pal_bg(pal_redforest);
				pal_spr(pal_sprites_greenforest);
				music_play(music_stage2); 
				break;
			case 2:
				pal_bg(pal_castle);
				pal_spr(pal_sprites_greenforest);
				music_play(music_stage3); 
				break;
			case 3:
				music_stop();
				break;
		}
	}

	levelIndex = index;
	roomIndex = 0;
	
	// load room
	unrle_vram(levels[levelIndex][roomIndex], 0x2000);
	unrle_vram(levels[levelIndex][roomIndex + 1], 0x2400);
	playerInit(spawnPoints[levelIndex * 2], spawnPoints[levelIndex * 2 + 1]);
	health = 6;
	hasKey = 0;
	pickedUpPowerups = 0x0000;
	processRoom();
	
	fadeIn();
}

void loadRoom(unsigned char index, unsigned char sx, unsigned char sy)
{
	fadeOut();
	oam_clear();
	
	roomIndex = index;
	unrle_vram(levels[levelIndex][roomIndex * 3], 0x2000);
	unrle_vram(levels[levelIndex][roomIndex * 3 + 1], 0x2400);
	playerInit(sx, sy);
	processRoom();
	
	fadeIn();
}

extern unsigned char playerPixelX;
extern unsigned char playerPixelY;
extern unsigned char weaponLevel;
extern unsigned char attacking;
void __fastcall__ levelUpdate(void)
{
	t0 = 52;
	for (i = 0; i < MAX_POWERUPS_PER_ROOM; ++i)
	{
		t3 = powerUpLocations[i * 2] & 0x80; // type
		x = powerUpLocations[i * 2] & 0x7F;
		if (x != 0)
		{
			t2 = x  - scrollX / 8;
			if (t2 < 32)
			{
				y = powerUpLocations[i * 2 + 1];
				t1 = ((playerPixelX + scrollX) / 8) - x;
				if (t1 > 0xFE ||
					t1 < 0x02)
				{
					t1 = ((playerPixelY + scrollY) / 8) - y;
					if (t1 > 0xFD ||
						t1 < 0x03)
					{
						powerUpLocations[i * 2] = 0;
						t0 = hideSpr(t0);
						if (t3) { if (health < 8) ++health; }
						else if (weaponLevel < 4) ++weaponLevel;
						attacking = 0;
						pickedUpPowerups |= (0x0001 << ((roomIndex * 2) + i));
						sfx_play(2, 0);
						continue;
					}
				}
				
				t1 = frameCounter % 24;
				if (t1 < 6)
					t1 = 1;
				else if (t1 < 12)
					t1 = 2;
				else if (t1 < 18)
					t1 = 1;
				else
					t1 = 0;
				t0 = oam_spr((x * 8) - scrollX, (y * 8) - scrollY + t1, t3 ? 0x0D : 0x04, 1, t0);
			}
			else
			{
				t0 = hideSpr(t0);
			}
		}
		else
		{
			t0 = hideSpr(t0);
		}
	}
	
	if (!hasKey && keyLocationX)
	{
		t1 = ((playerPixelX + scrollX) / 8) - keyLocationX;
		if (t1 > 0xFE ||
			t1 < 0x02)
		{
			t1 = ((playerPixelY + scrollY) / 8) - keyLocationY;
			if (t1 > 0xFD ||
				t1 < 0x03)
			{
				hasKey = 1;
				sfx_play(2, 0);
			}
		}	
		
		if ((keyLocationX - scrollX / 8) < 32)
		{
			t0 = oam_spr((keyLocationX * 8) - scrollX + 4, (keyLocationY * 8) - scrollY - 8, 0x0E, 1, t0);
		}
		else
		{
			t0 = hideSpr(t0);
		}
	}
	else
	{
			t0 = hideSpr(t0);
	}
}

extern unsigned char extraLives;
unsigned char checkDoor(void)
{
	#define doormap cp1
	doormap = levels[levelIndex][roomIndex * 3 + 2];
	for (i = 0; i < doormap[0]; ++i)
	{
		x = doormap[i * VARS_PER_DOORMAP_ENTRY + 1];
		y = doormap[i * VARS_PER_DOORMAP_ENTRY + 2];
		t0 = doormap[i * VARS_PER_DOORMAP_ENTRY + 3]; // destination
		
		if (t0 == 0xFF)
		{
			if (hasKey)
			{
				if (extraLives < 2) ++extraLives;
				sfx_play(6, 0);
				loadLevel(levelIndex + 1);
				return 1;
			}
		}
		else
		{
			if (((playerPixelX + scrollX) / 8 == x || (playerPixelX + scrollX) / 8 == x + 1) &&
				(playerPixelY + scrollY) / 8 == y)
			{
				x = doormap[i * VARS_PER_DOORMAP_ENTRY + 4]; // destination x
				y = doormap[i * VARS_PER_DOORMAP_ENTRY + 5]; // destination y
				sfx_play(6, 0);
				loadRoom(t0, x, y);
				return 1;
			}
		}
	}
	
	return 0;
	#undef doormap
}