static int playerVelX, playerVelY; // 8th pixels
static unsigned int playerX, playerY;
static unsigned char playerPixelX, playerPixelY;
static unsigned char playerState;
static unsigned char playerFlags;
static unsigned char animCounter;
static unsigned char animFrame;
const unsigned int scrollMargin = 120 * SUB_PIXELS;
static unsigned char weaponLevel;
#define MAX_WEAPON_LEVEL 4
static unsigned char attacking;
#define MAX_HEALTH 8
#define MAX_EXTRA_LIVES 2
static unsigned char health;
static unsigned char invincible;
static int attackX, attackY;
static unsigned char cooldown;
static unsigned char extraLives;

#define playerWorldX (playerX + (scrollX * SUB_PIXELS))
#define playerWorldY (playerY + (scrollY * SUB_PIXELS))

enum PLAYER_STATE
{
	GROUNDED = 0,
	JUMPING = 1, // multiple frames
	AIRBORN = 5,
};

enum PLAYER_FLAGS
{
	FACING_RIGHT = 0x01,
	JUMP_HELD = 0x02,
	ATTACK_HELD = 0x04,
};

signed char getRunSpeed(void)
{
	return 7 + weaponLevel;
}

void playerInit(unsigned char sx, unsigned char sy)
{
	scrollX = 0;
	playerX = (sx + 1) * 8 * SUB_PIXELS;
	playerY = sy * 8 * SUB_PIXELS - 1;
	while (scrollX < 0xFF && playerX > scrollMargin)
	{
		++scrollX;
		playerX -= SUB_PIXELS;
	}
	playerPixelX = playerX / SUB_PIXELS;
	playerPixelY = playerY / SUB_PIXELS;
	playerState = GROUNDED;
	playerVelX = playerVelY = 0;
	playerFlags = FACING_RIGHT;
	animCounter = animFrame = 0;
	hideSpr(hideSpr(hideSpr(hideSpr(hideSpr(hideSpr(24 * 4))))));
	scroll(scrollX, scrollY);
	invincible = 0;
	cooldown = 0;
}

void death()
{
	sfx_play(8, 0);
	if (extraLives == 0)
	{
		// go to title screen
		gameState = 0;
	}
	else
	{
		weaponLevel = 0;
		loadLevel(levelIndex);
		--extraLives;
	}
}

void hurt(unsigned char d)
{
	if (health <= d)
	{
		death();
	}
	else
	{
		health -= d;
		if (playerFlags & FACING_RIGHT) playerVelX = -SUB_PIXELS;
		else playerVelX = SUB_PIXELS;
		playerVelY = -SUB_PIXELS - SUB_PIXELS / 2;
		playerState = JUMPING;
		invincible = 128;
		sfx_play(7, 0);
	}
}

void __fastcall__ playerUpdate(void)
{
	if (invincible) --invincible;
	if (cooldown) --cooldown;
	
	// set velocities
	gamepad = pad_poll(0);
	if (playerState == GROUNDED)
	{
		// go into doors
		if (!attacking && gamepad & PAD_UP)
		{
			if (checkDoor())
				return;
		}
	
		// move right
		if (!attacking && gamepad & PAD_RIGHT)
		{
			if (playerVelX < getRunSpeed())
				++playerVelX;
				
			if (!(playerFlags & FACING_RIGHT))
			{
				if (playerVelX > 0)
					playerFlags |= FACING_RIGHT;
			}
		}
		else
		{
			if (playerVelX > 0)
				--playerVelX;
		}
		
		// move left
		if (!attacking && gamepad & PAD_LEFT)
		{
			if (playerVelX > -getRunSpeed())
				--playerVelX;
				
			if (playerFlags & FACING_RIGHT)
			{
				if (playerVelX < 0)
					playerFlags &= ~FACING_RIGHT;
			}
		}
		else
		{
			if (playerVelX < 0)
				++playerVelX;
		}
		
		// jump
		if (!(gamepad & PAD_A))
		{
			// remove the jump held flag when player lets go of the jump button
			// so we can jump again.
			playerFlags &= ~JUMP_HELD;
		}
		if (!(playerFlags & JUMP_HELD) && (gamepad & PAD_A))
		{
			// set jump button held
			playerFlags |= JUMP_HELD;
		
			// force turn
			if (gamepad & PAD_LEFT) playerVelX = -getRunSpeed();
			else if (gamepad & PAD_RIGHT) playerVelX = getRunSpeed();
			else playerVelX = 0;
		
			// force max speed on jump
			if (playerVelX > 0)
			{
				playerFlags |= FACING_RIGHT;
			}
			else if (playerVelX < 0)
			{
				playerFlags &= ~FACING_RIGHT;
			}
				
			playerVelY = -SUB_PIXELS * 2;
			playerState = JUMPING;
			sfx_play(3, 1);
		}
		else
		{			
			// gravity of 1 pixel to always check floor collision each frame
			playerVelY = SUB_PIXELS;
		}
	}
	else if (playerState < AIRBORN) // meaning we are jumping
	{
		// we don't want the gravity of air born directly after jumping,
		// this makes it wait for a couple of frames
		++playerState;
	}
	else if (playerState == AIRBORN)
	{
		if (playerVelY < (SUB_PIXELS * 2))
		{
			playerVelY += 1;
		}
	}
	// attack
	if (!(gamepad & PAD_B))
	{
		// remove the jump held flag when player lets go of the jump button
		// so we can jump again.
		playerFlags &= ~ATTACK_HELD;
	}
	if (!cooldown && !attacking && !(playerFlags & ATTACK_HELD) && (gamepad & PAD_B))
	{
		// set jump button held
		playerFlags |= ATTACK_HELD;
	
		// force turn
		if (gamepad & PAD_LEFT) playerFlags &= ~FACING_RIGHT;
		else if (gamepad & PAD_RIGHT) playerFlags |= FACING_RIGHT;
	
		attacking = 1;
	}
	
	// process horizontal velocity
	playerX += playerVelX;
	while (scrollX < 0xFF && playerX > scrollMargin)
	{
		++scrollX;
		playerX -= SUB_PIXELS;
	}
	while (scrollX > 0 && playerX < scrollMargin)
	{
		--scrollX;
		playerX += SUB_PIXELS;
	}
	playerPixelX = playerX / SUB_PIXELS;
	if (
		isSolid((playerPixelX + scrollX + 6) / 8, (playerPixelY + scrollY) / 8) ||
		isSolid((playerPixelX + scrollX - 6) / 8, (playerPixelY + scrollY) / 8)
		)
	{
		playerX -= playerVelX;
		playerPixelX = playerX / SUB_PIXELS;
	}
	
	// process vertical velocity
	playerY += playerVelY;
	if (playerY > 232 * SUB_PIXELS)
	{
		death();
		return;
	}
	playerPixelY = playerY / SUB_PIXELS;
	if (
		isSolid((playerPixelX + scrollX - 6) / 8, (playerPixelY + scrollY) / 8) ||
		isSolid((playerPixelX + scrollX + 6) / 8, (playerPixelY + scrollY) / 8)
		)
	{
		if (playerState == AIRBORN)
		{
			// we just landed
			playerState = GROUNDED;
			attacking = 0;
		}
		playerY -= playerVelY;
		playerPixelY = playerY / SUB_PIXELS;
		playerVelY = 0;
	}
	else
	{
		if (playerState == GROUNDED)
		{
			playerState = AIRBORN;
		}
	}
	

	if (attacking)
	{
		t1 = playerFlags & FACING_RIGHT;
		
		// these are used for collision testing
		if (t1)
			attackX = playerWorldX + 8 * SUB_PIXELS;
		else
			attackX = playerWorldX - 8 * SUB_PIXELS;
		attackY = playerWorldY - 12 * SUB_PIXELS;
	
		// process attacking
		t0 = 19 * 4;
		x = playerPixelX + (t1 ? 7 : -15);
		y = playerPixelY - 16;
		if (attacking >= (weaponLevel * 3) + 1)
		{
			if (attacking == (weaponLevel * 3) + 1)
			{
				sfx_play(5, 0);
			}
			t0 = oam_spr(t1 ? (x + weaponLevel * 8) : (x - weaponLevel * 8), y, weaponLevel ? 0x07 : 0x06, 1 | (t1 ? 0 : OAM_FLIP_H), t0);
		}
		else
		{
			t0 += 4;
		}
		
		if (attacking < (weaponLevel * 3) + 1 && (attacking % 3 == 1))
			sfx_play(4, 0);
		for (i = 0; i < 4; ++i)
		{
			if (i < weaponLevel && i < (attacking / 3) + 1)
				t0 = oam_spr(t1 ? (x + i * 8) : (x - i * 8), y, ((i + 1) < (attacking / 3) + 1) ? 0x05 : 0x07, 1, t0);
			else
				t0 = hideSpr(t0);
		}
		
		// update sprites when attacking
		if (t1)
		{
			// head
			t0 = oam_spr(playerPixelX - 8, playerPixelY - 24, 0x08, 1, 24 * 4);
			t0 = oam_spr(playerPixelX, playerPixelY - 24, 0x09, 1, t0);
			
			// body
			t0 = oam_spr(playerPixelX - 8, playerPixelY - 16, 0x1A, 0, t0);
			t0 = oam_spr(playerPixelX + 0, playerPixelY - 16, 0x1B, 0, t0);
			t0 = oam_spr(playerPixelX - 8, playerPixelY -  8, 0x2A, 0, t0);
			t0 = oam_spr(playerPixelX + 0, playerPixelY -  8, 0x2B, 0, t0);
		}
		else
		{
			// head
			t0 = oam_spr(playerPixelX - 8, playerPixelY - 24, 0x09, 1 | OAM_FLIP_H, 24 * 4);
			t0 = oam_spr(playerPixelX, playerPixelY - 24, 0x08, 1 | OAM_FLIP_H, t0);
			
			// body
			t0 = oam_spr(playerPixelX - 8, playerPixelY - 16, 0x1B + t1, OAM_FLIP_H, t0);
			t0 = oam_spr(playerPixelX + 0, playerPixelY - 16, 0x1A + t1, OAM_FLIP_H, t0);
			t0 = oam_spr(playerPixelX - 8, playerPixelY -  8, 0x2B + t1, OAM_FLIP_H, t0);
			t0 = oam_spr(playerPixelX + 0, playerPixelY -  8, 0x2A + t1, OAM_FLIP_H, t0);
		}
		
		if (++attacking == 16 + weaponLevel)
		{
			attacking = 0;
			cooldown = 12;
			hideSpr(19 * 4);
		}
	}
	else
	{
		hideSpr(hideSpr(hideSpr(hideSpr(hideSpr(19 * 4)))));
	
		// update sprites when not attacking
		if (playerFlags & FACING_RIGHT)
		{
			if (playerState == GROUNDED)
			{
				if (playerVelX > 0)
				{
					// count animation frames
					if (++animCounter == 8)
					{
						animCounter = 0;
						if (++animFrame == 4)
						{
							animFrame = 0;
						}
					}
					
					// head
					x = playerPixelX - 8;
					y = playerPixelY - 24;
					switch (animFrame)
					{
						case 0:
						case 2:  --x; ++y; break;
					}
					t0 = oam_spr(x, y, 0x08, 1, 24 * 4);
					t0 = oam_spr(x + 8, y, 0x09, 1, t0);
					
					// body
					t1 = 2 * animFrame;
					t0 = oam_spr(playerPixelX - 8, playerPixelY - 16, 0x12 + t1, 0, t0);
					t0 = oam_spr(playerPixelX + 0, playerPixelY - 16, 0x13 + t1, 0, t0);
					t0 = oam_spr(playerPixelX - 8, playerPixelY -  8, 0x22 + t1, 0, t0);
					t0 = oam_spr(playerPixelX + 0, playerPixelY -  8, 0x23 + t1, 0, t0);
				}
				else
				{
					goto __NORMAL__RIGHT;
				}
			}
			else
			{
				__NORMAL__RIGHT:
				animFrame = animCounter = 0;
				t0 = oam_spr(playerPixelX - 8, playerPixelY - 24, 0x08, 1, 24 * 4);
				t0 = oam_spr(playerPixelX + 0, playerPixelY - 24, 0x09, 1, t0);
				t0 = oam_spr(playerPixelX - 8, playerPixelY - 16, 0x10, 0, t0);
				t0 = oam_spr(playerPixelX + 0, playerPixelY - 16, 0x11, 0, t0);
				t0 = oam_spr(playerPixelX - 8, playerPixelY -  8, 0x20, 0, t0);
				t0 = oam_spr(playerPixelX + 0, playerPixelY -  8, 0x21, 0, t0);
			}
		}
		else
		{
			if (playerState == GROUNDED)
			{
				if (playerVelX < 0)
				{
					// count animation frames
					if (++animCounter == 8)
					{
						animCounter = 0;
						if (++animFrame == 4)
						{
							animFrame = 0;
						}
					}
					
					// head
					x = playerPixelX - 8;
					y = playerPixelY - 24;
					switch (animFrame)
					{
						case 0:
						case 2:  ++x; ++y; break;
					}
					t0 = oam_spr(x, y, 0x09, 1 | OAM_FLIP_H, 24 * 4);
					t0 = oam_spr(x + 8, y, 0x08, 1 | OAM_FLIP_H, t0);
					
					// body
					t1 = 2 * animFrame;
					t0 = oam_spr(playerPixelX - 8, playerPixelY - 16, 0x13 + t1, OAM_FLIP_H, t0);
					t0 = oam_spr(playerPixelX + 0, playerPixelY - 16, 0x12 + t1, OAM_FLIP_H, t0);
					t0 = oam_spr(playerPixelX - 8, playerPixelY -  8, 0x23 + t1, OAM_FLIP_H, t0);
					t0 = oam_spr(playerPixelX + 0, playerPixelY -  8, 0x22 + t1, OAM_FLIP_H, t0);
				}
				else
				{
					goto __NORMAL__LEFT;
				}
			}
			else
			{
				__NORMAL__LEFT:
				animFrame = animCounter = 0;
				t0 = oam_spr(playerPixelX - 8, playerPixelY - 24, 0x09, 1 | OAM_FLIP_H, 24 * 4);
				t0 = oam_spr(playerPixelX + 0, playerPixelY - 24, 0x08, 1 | OAM_FLIP_H, t0);
				t0 = oam_spr(playerPixelX - 8, playerPixelY - 16, 0x11, OAM_FLIP_H, t0);
				t0 = oam_spr(playerPixelX + 0, playerPixelY - 16, 0x10, OAM_FLIP_H, t0);
				t0 = oam_spr(playerPixelX - 8, playerPixelY -  8, 0x21, OAM_FLIP_H, t0);
				t0 = oam_spr(playerPixelX + 0, playerPixelY -  8, 0x20, OAM_FLIP_H, t0);
			}
		}
	}
}