//---------------------------------------------------------------------------

#include <vcl.h>
#include <stdio.h>
#pragma hdrstop

#include "UnitMain.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TFormMain *FormMain;


unsigned char *nes;
FILETIME nesLastModified;



void __fastcall TFormMain::DrawPRG(int x,int y,unsigned char *data,int fill)
{
	int i,j;
	unsigned char *dst;

	for(i=0;i<256;i++)
	{
		dst=(unsigned char*)ImagePRG->Picture->Bitmap->ScanLine[i];

		for(j=0;j<64;j++) *dst++=(*data++==fill)?0:1;
	}

	PaintBoxInfo->Canvas->Rectangle(x,y,x+66,y+258);
	PaintBoxInfo->Canvas->Draw(x+1,y+1,ImagePRG->Picture->Bitmap);
}



void __fastcall TFormMain::DrawCHR(int x,int y,unsigned char *data)
{
	int i,j;
	unsigned char *dst;

	for(i=0;i<128;i++)
	{
		dst=(unsigned char*)ImageCHR->Picture->Bitmap->ScanLine[i];

		for(j=0;j<64;j++) *dst++=*data++?1:0;
	}

	PaintBoxInfo->Canvas->Rectangle(x,y,x+66,y+130);
	PaintBoxInfo->Canvas->Draw(x+1,y+1,ImageCHR->Picture->Bitmap);
}



bool __fastcall TFormMain::LoadNES(AnsiString name)
{
	WIN32_FILE_ATTRIBUTE_DATA fileAttrData={0};
	FILE *file;
	unsigned char header[16];
	int size;

	file=fopen(name.c_str(),"rb");

	if(!file) return false;

	fseek(file,0,SEEK_END);
	size=ftell(file);
	fseek(file,0,SEEK_SET);

	if(size<16+16384+8192)//file is too small
	{
		fclose(file);
		return false;
	}

	if(nes) free(nes);

	fread(header,16,1,file);

	if(memcmp(header,"NES",3))//not a NES ROM
	{
		fclose(file);
		return false;
	}

	nes=(unsigned char*)malloc(size);
	fseek(file,0,SEEK_SET);
	fread(nes,size,1,file);
	fclose(file);

	GetFileAttributesEx(name.c_str(),GetFileExInfoStandard,&fileAttrData);
	nesLastModified=fileAttrData.ftLastWriteTime;

	return true;
}



void __fastcall TFormMain::ShowNES(void)
{
	int lens[256];
	int i,j,x,y,prg,chr,mapper,off,fill,hgt,len,prev,max;

	PaintBoxInfo->Canvas->Rectangle(0,0,PaintBoxInfo->Width,PaintBoxInfo->Height);

	if(!nes)
	{
		PaintBoxInfo->Canvas->TextOut(10,10,"No file loaded");
		return;
	}

	prg=nes[4];
	chr=nes[5];
	mapper=((nes[6]>>4)&0x0f)|(nes[7]&0xf0);

	for(i=0;i<256;i++) lens[i]=0;

	off=16;

	for(i=0;i<prg;i++)
	{
		len=0;
		prev=nes[off++];

		for(j=1;j<16384;j++)
		{
			if(nes[off]==prev)
			{
				len++;
			}
			else
			{
				if(len>=8) lens[prev]=len;
				prev=nes[off];
				len=0;
			}

			off++;
		}
	}

	max=-1;
	fill=-1;

	for(i=0;i<256;i++)
	{
		if(lens[i]>max)
		{
			max=lens[i];
			fill=i;
		}
	}

	PaintBoxInfo->Canvas->TextOut(10,10,OpenDialogNES->FileName+"  Mapper #"+IntToStr(mapper)+"  "+IntToStr(prg)+"xPRG  "+IntToStr(chr)+"xCHR  Empty is $"+IntToHex(fill,2));

	off=16;
	x=10;
	y=50;
	hgt=256+40;

	for(i=0;i<prg;i++)
	{
		PaintBoxInfo->Canvas->TextOut(x,y,"PRG $"+IntToHex(i,2));
		DrawPRG(x,y+20,&nes[off],fill);

		off+=16384;
		x+=64+10;

		if(x+64+10>=PaintBoxInfo->Width)
		{
			x=10;
			y+=hgt;
			if(i==prg-1) hgt=128+40;
		}
	}

	for(i=0;i<chr;i++)
	{
		PaintBoxInfo->Canvas->TextOut(x,y,"CHR $"+IntToHex(i,2));
		DrawCHR(x,y+20,&nes[off]);

		off+=8192;
		x+=64+10;

		if(x+64+10>=PaintBoxInfo->Width)
		{
			x=10;
			y+=hgt;
			hgt=128+40;
		}
	}
}



void __fastcall TFormMain::DropFilesHandler(TWMDropFiles &Message)
{
	HDROP hd;
	char buf[MAX_PATH];

	hd=(HDROP)Message.Drop;

	if(DragQueryFile(hd,-1,NULL,NULL))
	{
		DragQueryFile(hd,0,buf,sizeof(buf));
		OpenDialogNES->FileName=buf;
		LoadNES(buf);
		ShowNES();
	}

	DragFinish(hd);
}



//---------------------------------------------------------------------------
__fastcall TFormMain::TFormMain(TComponent* Owner)
: TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TFormMain::MOpenClick(TObject *Sender)
{
	if(OpenDialogNES->Execute())
	{
		LoadNES(OpenDialogNES->FileName);
		ShowNES();
	}
}
//---------------------------------------------------------------------------
void __fastcall TFormMain::FormCreate(TObject *Sender)
{
	nes=NULL;
	nesLastModified.dwLowDateTime=0;
	nesLastModified.dwHighDateTime=0;

	if(ParamStr(1)!="")
	{
		OpenDialogNES->FileName=ParamStr(1);
		LoadNES(ParamStr(1));
	}

	DragAcceptFiles(Handle,true);
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormDestroy(TObject *Sender)
{
	if(nes) free(nes);
}
//---------------------------------------------------------------------------



void __fastcall TFormMain::MReloadClick(TObject *Sender)
{
	LoadNES(OpenDialogNES->FileName);
	ShowNES();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::MSaveReportClick(TObject *Sender)
{
	Graphics::TBitmap *img;

	if(SavePictureDialogReport->Execute())
	{
		img=new Graphics::TBitmap();
		img->Width=PaintBoxInfo->Width;
		img->Height=PaintBoxInfo->Height;

		ShowNES();
		img->Canvas->CopyRect(Rect(0,0,PaintBoxInfo->Width,PaintBoxInfo->Height),PaintBoxInfo->Canvas,Rect(0,0,PaintBoxInfo->Width,PaintBoxInfo->Height));

		img->SaveToFile(SavePictureDialogReport->FileName);

		delete img;
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormPaint(TObject *Sender)
{
	ShowNES();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::TimerWatchTimer(TObject *Sender)
{
	WIN32_FILE_ATTRIBUTE_DATA fileAttrData={0};

	if(GetFileAttributesEx(OpenDialogNES->FileName.c_str(),GetFileExInfoStandard,&fileAttrData))
	{
		if(CompareFileTime((const FILETIME*)&fileAttrData.ftLastWriteTime,(const FILETIME*)&nesLastModified))
		{
			LoadNES(OpenDialogNES->FileName.c_str());
			ShowNES();
		}
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::MWatchClick(TObject *Sender)
{
	if(!TimerWatch->Enabled)
	{
		MWatch->Caption="Unwatch";
		TimerWatch->Enabled=true;
	}
	else
	{
		MWatch->Caption="Watch";
		TimerWatch->Enabled=false;
	}
}
//---------------------------------------------------------------------------

